@@ -23,6 +23,7 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -866,8 +867,7 @@ dump_lookup_subtable_gsub (int indent, int index, unsigned type,
 	      subtable->u.extension1.ExtensionOffset);
       dump_lookup_subtable_gsub (indent, index, 
 				 subtable->u.extension1.ExtensionLookupType,
-				 (subtable
-				  + subtable->u.extension1.ExtensionOffset));
+				 subtable->u.extension1.ExtensionSubtable);
       break;
 
     case 8:
@@ -996,7 +996,38 @@ dump_lookup_subtable_gpos (int indent, int index, unsigned type,
     case 5:
       if (subtable->Format == 1)
 	{
-	  printf (" not-yet-supported");
+	  OTF_GPOS_MarkLig1 *mark_lig1 = &subtable->u.mark_lig1;
+	  int i, j, k;
+
+	  free (dump_coverage (indent, "Mark", &subtable->Coverage));
+	  free (dump_coverage (indent, "Ligature",
+			       &mark_lig1->LigatureCoverage));
+	  IPRINT ("(ClassCount %d)", mark_lig1->ClassCount);
+	  dump_mark_array (indent, &mark_lig1->MarkArray);
+	  IPRINT ("(LigatureArray (%d)",
+		  mark_lig1->LigatureArray.LigatureCount);
+	  indent++;
+	  for (i = 0; i < mark_lig1->LigatureArray.LigatureCount; i++)
+	    {
+	      OTF_LigatureAttach *attach
+		= mark_lig1->LigatureArray.LigatureAttach + i;
+
+	      IPRINT ("(LigatureAttach (%d)", attach->ComponentCount);
+	      indent++;
+	      for (j = 0; j < attach->ComponentCount; j++)
+		{
+		  OTF_ComponentRecord *rec = attach->ComponentRecord + j;
+
+		  IPRINT ("(LigatureAnchor (%d)", mark_lig1->ClassCount);
+		  for (k = 0; k < mark_lig1->ClassCount; k++)
+		    if (rec->LigatureAnchor[k].AnchorFormat)
+		      dump_anchor (indent + 1, rec->LigatureAnchor + k);
+		  printf (")");
+		}
+	      printf (")");
+	      indent--;
+	    }
+	  printf (")");
 	}
       else
 	printf (" invalid");
@@ -1090,6 +1121,21 @@ dump_lookup_subtable_gpos (int indent, int index, unsigned type,
       else
 	printf (" invalid");
       break;
+
+    case 9:
+      if (subtable->Format == 1)
+	{
+	  IPRINT ("(ExtensionLookupType %d)",
+		  subtable->u.extension1.ExtensionLookupType);
+	  IPRINT ("(ExtensionOffset %d)",
+		  subtable->u.extension1.ExtensionOffset);
+	  dump_lookup_subtable_gpos
+	    (indent, index, 
+	     subtable->u.extension1.ExtensionLookupType,
+	     subtable->u.extension1.ExtensionSubtable);
+	}
+      else
+	printf (" invalid");
     }
   printf (")");
 }
@@ -1310,6 +1356,22 @@ dump_cmap_table (int indent, OTF_cmap *cmap)
 	    printf (")");
 	  }
 	  break;
+
+	case 12:
+	  {
+	    OTF_EncodingSubtable12 *sub12
+	      = cmap->EncodingRecord[i].subtable.f.f12;
+	    int j;
+
+	    for (j = 0; j < sub12->nGroups; j++)
+	      {
+		IPRINT ("(Group (#x%X) (startChar #x%04X) (endChar #x%04X) (startGlyphID #x%X))",
+			j,
+			sub12->Groups[j].startCharCode,
+			sub12->Groups[j].endCharCode,
+			sub12->Groups[j].startGlyphID);
+	      }
+	  }
 	}
 
       indent -= 2;
