@@ -24,8 +24,15 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
 #ifndef _OTF_H_
 #define _OTF_H_
 
-/* Version number of this library.  */
-#define LIBOTF_VERSION "0.9"
+/* Version name of this library.  */
+#define LIBOTF_VERSION "0.9.3"
+
+/* Major version number.  */
+#define LIBOTF_MAJOR_VERSION 0
+/* Minor version number.  */
+#define LIBOTF_MINOR_VERSION 9
+/* Release (i.e. patch level) number.  */
+#define LIBOTF_RELEASE_NUMBER 3
 
 /***
     Table of contents:
@@ -56,6 +63,8 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
     (3-4) OTF_drive_gsub()
     (3-5) OTF_drive_gpos()
     (3-6) OTF_drive_tables()
+    (3-7) OTF_get_unicode()
+    (3-8) OTF_drive_gsub_alternate()
 
     (4) API for error handling
     (4-1) Error codes
@@ -104,6 +113,10 @@ typedef struct
   int nameID;
   int length;
   int offset;
+
+  /* If nonzero, NAME is an ASCII string.  */
+  int ascii;
+  unsigned char *name;
 } OTF_NameRecord;
 
 #define OTF_max_nameID 23
@@ -135,9 +148,11 @@ typedef struct
 
 typedef struct
 {
-  unsigned subHeaderKeys[256];
+  unsigned short subHeaderKeys[256];
+  int subHeaderCount;
   OTF_cmapSubHeader *subHeaders;
-  unsigned *glyphIndexArray;
+  int glyphIndexCount;
+  OTF_GlyphID *glyphIndexArray;
 } OTF_EncodingSubtable2;
 
 typedef struct
@@ -449,11 +464,14 @@ enum OTF_LookupFlagBit
     OTF_RightToLeft = 0x0001,
     OTF_IgnoreBaseGlyphs = 0x0002,
     OTF_IgnoreLigatures = 0x0004,
-    OTF_IgnoreMarks = 0x8000,
+    OTF_IgnoreMarks = 0x0008,
     OTF_Reserved = 0x00F0,
     OTF_MarkAttachmentType = 0xFF00
   };
 
+#define OTF_LookupFlagIgnoreMask \
+  (OTF_IgnoreBaseGlyphs | OTF_IgnoreLigatures | OTF_IgnoreMarks)
+
 typedef struct
 {
   OTF_Offset offset;
@@ -1131,6 +1149,9 @@ typedef struct
      drawn at the normal position.  */
   int positioning_type;
   union {
+    struct {
+      int from, to;
+    } index;
     struct {
       enum OTF_ValueFormat format;
       OTF_ValueRecord *value;
@@ -1167,7 +1188,7 @@ typedef struct
   /* How many glyphs are allocated at the memory pointed by the member
       <glyphs>.  */
   int size;
-  /* How many glyphs contains valid information.  */
+  /* How many glyphs contain valid information.  */
   int used;
   /* Array of glyphs.  It must be allocated by malloc().  The
      functions otf_drive_XXX() may reallocate it and increase the
@@ -1179,14 +1200,27 @@ typedef struct
 /*** (3-2) OTF_drive_cmap() */
 
 /***
-    Process glyph string by cmap table.
+    Process glyph string by Unicode-based cmap table.
 
-    The OTF_drive_cmap() function looks up the cmap table of OpenType
-    font $OTF, and setup the member <glyhph_id> of all glhphs in the
-    glyph string $GSTRING if the value of the member is not zero.  */
+    The OTF_drive_cmap() function looks up a Unicode-based cmap table
+    of OpenType font $OTF, and setup the member <glyhph_id> of all
+    glhphs in the glyph string $GSTRING if the value of the member is
+    not zero.  */
 
 extern int OTF_drive_cmap (OTF *otf, OTF_GlyphString *gstring);
 
+/***
+    Process glyph string by a specific cmap table.
+
+    The OTF_drive_cmap2() function looks up a cmap table (whose
+    Platform-ID is $PLATFORM_ID an Encoding-ID is $ENCODING_ID) of
+    OpenType font $OTF, and setup the member <glyhph_id> of all glhphs
+    in the glyph string $GSTRING if the value of the member is not
+    zero.  */
+
+extern int OTF_drive_cmap2 (OTF *otf, OTF_GlyphString *gstring,
+			    int platform_id, int encoding_id);
+
 /*** (3-3) OTF_drive_gdef() */
 
 /***
@@ -1208,7 +1242,9 @@ extern int OTF_drive_gdef (OTF *otf, OTF_GlyphString *gstring);
     font $OTF, and by using features the font has for script $SCRIPT
     and language system $LANGSYS, update member <glyphs> of the glyph
     string $GSTRING.  It may substitute, delete, insert glyphs in that
-    array.  $FEATURES is a list of features to apply.  */
+    array.  $FEATURES is a list of features to apply.  This doesn't
+    perform a lookup of type 3 (Alternate Substitution).  For that,
+    use OTF_drive_gsub_alternate().  */
 
 extern int OTF_drive_gsub (OTF *otf, OTF_GlyphString *gstring,
 			   char *script, char *language, char *features);
@@ -1249,6 +1285,18 @@ extern int OTF_drive_tables (OTF *otf, OTF_GlyphString *gstring,
 
 extern int OTF_get_unicode (OTF *otf, OTF_GlyphID code);
 
+/*** (3-8) OTF_drive_gsub_alternate() */
+
+/***
+    Find alternate glyphs.
+
+    This is like OTF_drive_gsub(), but perform only a lookup of type 3
+    (Alternate Substituion).  */
+
+extern int OTF_drive_gsub_alternate (OTF *otf, OTF_GlyphString *gstring,
+				     char *script, char *language,
+				     char *features);
+
 /*** (4) API for error handling ***/
 
 /*** (4-1) Error codes ***/
